structure ElaborationTop :> ELABORATION_TOP =
  struct
    structure Elaborate :> ELABORATE = Elaborate

    fun process_program parse input =
        let val elprog = parse input
            val (sbnds, sdecs) = Elaborate.elab_program [] elprog
            val layout =
                Layout.align
                    [ Layout.seq [Layout.str "  |- ",
                                  ELPrint.program_to_layout elprog],
                      Layout.seq [Layout.str " ~~> ",
                                  ILPrint.sbnds_to_layout sbnds],
                      Layout.seq [Layout.str "   : ",
                                  ILPrint.sdecs_to_layout sdecs] ]
            (* weird... it seems to go over if you allow the default of 80 *)
            val s = Layout.tostringex 77 layout
        in
            print (s ^ "\n")
        (*
        in
            Layout.print (layout, print);
            print "\n"
        *)
        end
        handle
            Parser.Parse (pos, string) =>
                print ("Syntax error: at " ^ Pos.toString pos ^ ": " ^ string
                        ^ "\n")
          | Tokenize.Lex (pos, string) =>
                print ("Lexing error: at " ^ Pos.toString pos ^ ": " ^ string
                        ^ "\n")
          | Reparser.Reparse (pos, string) =>
                print ("Reparse error: at " ^ Pos.toString pos ^ ": " ^ string
                        ^ "\n")
          | Elaborate.Elab string =>
                print ("Elaboration error: " ^ string ^ "\n")
          | Elaborate.Error string =>
                print ("Elaborator internal error: " ^ string ^ "\n")
          | ILStatic.Check string =>
                print ("IL type error: " ^ string ^ "\n")
          | ILStatic.Error string =>
                print ("IL internal error: " ^ string ^ "\n")

    fun elab_string str = process_program Reparser.parse_string str
    fun elab_file file = process_program Reparser.parse_file file
  end
