structure ILStatic :> ILSTATIC =
  struct
    open IL

    type context = IL.dec list
    type tcontext = (IL.label * IL.dec) list

    exception Equiv
    exception Occurs
    exception Subkind
    exception Subsig
    exception Check of string
    exception Error of string

    fun check b f = if b then () else f ()

    fun nth (l, n) = SOME (List.nth (l,n)) handle Subscript => NONE


    (*** context/signature lookup ***)
    (* XXX publicize these? *)

    (* NB: newest sdec at the front of the list *)
    fun extend ctx dec = dec::ctx

    fun extend_many ctx decs = List.revAppend (decs, ctx)

    fun list_findSome pfun l =
        case l of
            [] => NONE
          | x::xs =>
                (case pfun x of
                    NONE => list_findSome pfun xs
                  | SOME y => SOME y)

    fun lookup_convar ctx convar =
        list_findSome
            (fn DecConKind (cx, k) =>
                    if Variable.eq (convar, cx) then SOME k else NONE
              | _ => NONE)
            ctx

    fun lookup_expvar ctx expvar =
        list_findSome
            (fn DecExpCon (x, c) =>
                    if Variable.eq (expvar, x) then SOME c else NONE
              | _ => NONE)
            ctx

    fun lookup_modvar ctx modvar =
        list_findSome
            (fn DecModSig (mx, sg) =>
                    if Variable.eq (modvar, mx) then SOME sg else NONE
              | _ => NONE)
            ctx


    (*** sdecs/sbnds lookup, that makes sure the output is externally valid
         by substituting away any references to variables bound earlier ***)

    (* XXX way to generalize these?  copy-paste = :( *)

    fun sdecs_lookup_conlab modv sdecs lab =
        case sdecs of
            [] => NONE
          | (l, DecConKind (cx, k))::sds =>
                if l = lab then
                    SOME k
                else
                    (*
                    sdecs_lookup_conlab modv sds lab >>= (fn k =>
                    subst_con_kind (ConModProj (modv, l)) cx k)
                    *)
                    Option.map (subst_con_kind (ConModProj (modv, l)) cx)
                               (sdecs_lookup_conlab modv sds lab)
          | (l, DecModSig (mx, sg))::sds =>
                Option.map (subst_mod_kind (ModModProj (modv, l)) mx)
                           (sdecs_lookup_conlab modv sds lab)
          | (l, DecExpCon (ex, c))::sds =>
                (* is this substitution really necessary?  phase distinction,
                   mumble mumble... *)
                Option.map (subst_exp_kind (ExModProj (modv, l)) ex)
                           (sdecs_lookup_conlab modv sds lab)

    fun sdecs_lookup_modlab modv sdecs lab =
        case sdecs of
            [] => NONE
          | (l, DecConKind (cx, k))::sds =>
                Option.map (subst_con_signat (ConModProj (modv, l)) cx)
                           (sdecs_lookup_modlab modv sds lab)
          | (l, DecModSig (mx, sg))::sds =>
                if l = lab then
                    SOME sg
                else
                    Option.map (subst_mod_signat (ModModProj (modv, l)) mx)
                               (sdecs_lookup_modlab modv sds lab)
          | (l, DecExpCon (ex, c))::sds =>
                (* is this substitution really necessary?  phase distinction,
                   mumble mumble... *)
                Option.map (subst_exp_signat (ExModProj (modv, l)) ex)
                           (sdecs_lookup_modlab modv sds lab)

    fun sdecs_lookup_explab modv sdecs lab =
        case sdecs of
            [] => NONE
          | (l, DecConKind (cx, k))::sds =>
                Option.map (subst_con_con (ConModProj (modv, l)) cx)
                           (sdecs_lookup_explab modv sds lab)
          | (l, DecModSig (mx, sg))::sds =>
                Option.map (subst_mod_con (ModModProj (modv, l)) mx)
                           (sdecs_lookup_explab modv sds lab)
          | (l, DecExpCon (ex, c))::sds =>
                if l = lab then
                    SOME c
                else
                    (* is this substitution really necessary?  phase
                       distinction, mumble mumble... *)
                    Option.map (subst_exp_con (ExModProj (modv, l)) ex)
                               (sdecs_lookup_explab modv sds lab)


    fun sbnds_lookup_conlab modv sbnds lab =
        case sbnds of
            [] => NONE
          | (l, BndCon (cx, c))::sbs =>
                if l = lab then
                    SOME c
                else
                    Option.map (subst_con_con (ConModProj (modv, l)) cx)
                               (sbnds_lookup_conlab modv sbs lab)
          | (l, BndMod (mx, m))::sbs =>
                Option.map (subst_mod_con (ModModProj (modv, l)) mx)
                           (sbnds_lookup_conlab modv sbs lab)
          | (l, BndExp (ex, e))::sbs =>
                (* is this substitution really necessary?  phase distinction,
                   mumble mumble... *)
                Option.map (subst_exp_con (ExModProj (modv, l)) ex)
                           (sbnds_lookup_conlab modv sbs lab)

    
    (*** module values and selfification ***)
    fun module_value module =
        case module of
            ModVar _ => true
          | ModStruct sbnds => sbnds_value sbnds
          | ModModProj (m, _) => module_value m
          | ModConstrain _ => false

    and sbnds_value sbnds =
        case sbnds of
            [] => true
          | sb::sbs => sbnd_value sb andalso sbnds_value sbs

    and sbnd_value (l, bnd) =
        case bnd of
            BndExp (ex, e) => exp_value e
          | BndCon _ => true (* XXX should we check modv.lab cons? *)
          | BndMod (mx, m) => module_value m

    and exp_value exp =
        case exp of
            ExVar ex => true
          | ExUnit => true
          | ExBool _ => true
          | ExIf _ => false
          | ExLit _ => true
          | ExUnOp _ => false
          | ExBinOp _ => false
          | ExPair (e1, e2) => exp_value e1 andalso exp_value e2
          | ExProj1 _ => false
          | ExProj2 _ => false
          | ExNil _ => true
          | ExCons (e1, e2) => exp_value e1 andalso exp_value e2
          | ExCase _ => false
          | ExFix _ => true
          | ExApp _ => false
          | ExConAbs (cx, k, e) => exp_value e (* XXX value restriction? *)
          | ExConApp (e, c) => exp_value e
          | ExModLet _ => false
          | ExModProj (m, l) => module_value m

    fun selfify modv sdecs =
        case sdecs of
            [] => []
          | (l, DecExpCon (ex, c)) :: sds =>
                (l, DecExpCon (ex, c))
                :: selfify modv sds
          | (l, DecConKind (cx, k)) :: sds =>
                (l, DecConKind (cx, labelledKndSing (ConModProj (modv, l)) k))
                :: selfify modv sds
          | (l, DecModSig (mx, SigStruct sds')) :: sds =>
                let val sds' = selfify (ModModProj (modv, l)) sds'
                in
                    (l, DecModSig (mx, SigStruct sds'))
                    :: selfify modv sds
                end


    (*** unification proper (used in equivalence below) ***)

    fun unify ebr con =
        case !ebr of
            EvBound _ => raise Error "bound Evar not eliminated by whnf"
          | EvFree (condeps, moddeps) =>
                (check_unifiable_con (ConVarSet.empty, ModVarSet.empty)
                                     ebr (condeps, moddeps) con;
                 ebr := EvBound con)

    (* checks that
            (a) the Evar with id doesn't occur in con (the "occurs check"), and
            (b) that any free constructor or module variables in con appear in
                condeps or moddeps, respectively.
       if both conditions hold, then `ConEvar (EvFree (condeps, moddeps))'
       is unifiable with con.  (keeps track along the way of bound constructor
       variables in bcvs and bound module variables in bmvs.) *)
    and check_unifiable_con (bcvs, bmvs) ebref (condeps, moddeps) con =
        (* NB: implemented in terms of IL.map_*, but never uses the return
           result, so the functions just return dummy values *)
        map_con (fn cx =>
                    if ConVarSet.member (bcvs, cx) orelse
                       ConVarSet.member (condeps, cx)
                    then
                        con
                    else
                        raise Occurs)
                (fn ebr =>
                    case !ebr of
                        EvBound c =>
                            check_unifiable_con (bcvs, bmvs)
                                                ebref (condeps, moddeps) c
                      | EvFree (cd, md) =>
                            if ebr = ebref then
                              raise Occurs
                            else
                              let val cd' = ConVarSet.intersection (condeps, cd)
                                  val md' = ModVarSet.intersection (moddeps, md)
                              in
                                  ebr := EvFree (cd', md');
                                  con
                              end)
                (fn (cx, c) =>
                    (check_unifiable_con (ConVarSet.add (bcvs, cx), bmvs)
                                         ebref (condeps, moddeps) c;
                     (cx, c)))
                (check_unifiable_kind (bcvs, bmvs) ebref (condeps, moddeps))
                (check_unifiable_module (bcvs, bmvs) ebref (condeps, moddeps))
                con

    and check_unifiable_kind (bcvs, bmvs) ebref (condeps, moddeps) knd =
        map_kind (fn (cx, k) =>
                    (check_unifiable_kind (ConVarSet.add (bcvs, cx), bmvs)
                                          ebref (condeps, moddeps) k;
                     (cx, k)))
                 (check_unifiable_con (bcvs, bmvs) ebref (condeps, moddeps))
                 knd

    and check_unifiable_module (bcvs, bmvs) ebref (condeps, moddeps) module =
        map_module (fn mx =>
                        if ModVarSet.member (bmvs, mx) orelse
                           ModVarSet.member (moddeps, mx)
                        then
                            module
                        else
                            raise Occurs)
                   (check_unifiable_sbnds (bcvs, bmvs) ebref (condeps, moddeps))
                   (check_unifiable_signat (bcvs, bmvs) ebref (condeps, moddeps))
                   module

    and check_unifiable_sbnds (bcvs, bmvs) ebref (condeps, moddeps) sbnds =
        map_sbnds (fn (ex, sbs) =>
                        (check_unifiable_sbnds (bcvs, bmvs)
                                               ebref (condeps, moddeps) sbs;
                         (ex, sbs)))
                  (fn (cx, sbs) =>
                        (check_unifiable_sbnds (ConVarSet.add (bcvs, cx), bmvs)
                                               ebref (condeps, moddeps) sbs;
                         (cx, sbs)))
                  (fn (mx, sbs) =>
                        (check_unifiable_sbnds (bcvs, ModVarSet.add (bmvs, mx))
                                               ebref (condeps, moddeps) sbs;
                         (mx, sbs)))
                  (check_unifiable_exp (bcvs, bmvs) ebref (condeps, moddeps))
                  (check_unifiable_con (bcvs, bmvs) ebref (condeps, moddeps))
                  (check_unifiable_module (bcvs, bmvs) ebref (condeps, moddeps))
                  sbnds

    and check_unifiable_signat (bcvs, bmvs) ebref (condeps, moddeps) signat =
        map_signat (check_unifiable_sdecs (bcvs, bmvs) ebref (condeps, moddeps))
                   signat

    and check_unifiable_sdecs (bcvs, bmvs) ebref (condeps, moddeps) sdecs =
        map_sdecs (fn (ex, sds) =>
                        (check_unifiable_sdecs (bcvs, bmvs)
                                               ebref (condeps, moddeps) sds;
                         (ex, sds)))
                  (fn (cx, sds) =>
                        (check_unifiable_sdecs (ConVarSet.add (bcvs, cx), bmvs)
                                               ebref (condeps, moddeps) sds;
                         (cx, sds)))
                  (fn (mx, sds) =>
                        (check_unifiable_sdecs (bcvs, ModVarSet.add (bmvs, mx))
                                               ebref (condeps, moddeps) sds;
                         (mx, sds)))
                  (check_unifiable_con (bcvs, bmvs) ebref (condeps, moddeps))
                  (check_unifiable_kind (bcvs, bmvs) ebref (condeps, moddeps))
                  (check_unifiable_signat (bcvs, bmvs) ebref (condeps, moddeps))
                  sdecs

    and check_unifiable_exp (bcvs, bmvs) ebref (condeps, moddeps) exp =
        map_exp (fn ex => exp) (* no checking to be done for a expvar *)
                (fn (cx, e) =>
                    (check_unifiable_exp (ConVarSet.add (bcvs, cx), bmvs)
                                         ebref (condeps, moddeps) e;
                     (cx, e)))
                (fn (mx, e) =>
                    (check_unifiable_exp (bcvs, ModVarSet.add (bmvs, mx))
                                         ebref (condeps, moddeps) e;
                     (mx, e)))
                (check_unifiable_con (bcvs, bmvs) ebref (condeps, moddeps))
                (check_unifiable_kind (bcvs, bmvs) ebref (condeps, moddeps))
                (check_unifiable_module (bcvs, bmvs) ebref (condeps, moddeps))
                exp

    
    (*** static semantics ***)

    (* precondition: module is a path *)
    fun extract_natural_sig ctx module =
        case module of
            ModVar mx =>
                (case lookup_modvar ctx mx of
                    NONE => raise Check
                            ("extract_natural_sig: unbound modvar, "
                            ^ Variable.toString mx)
                  | SOME sg => sg
                  (*
                  | SOME _ =>
                        (* XXX is the natural sig ever a functor?  we only
                           care about natural sigs because of the modv.lab
                           constructor form, and that modv better not be a
                           functor application, even of a variable... *)
                        raise Error "extract_natural_sig: found functor sig"
                  *)
                  )
          | ModModProj (modv, lab) =>
                let val SigStruct sdecs = extract_natural_sig ctx modv
                in
                    case sdecs_lookup_modlab modv sdecs lab of
                        SOME sg => sg
                      (* | SOME _ => raise Error "extract_natural_sig: bad sig" *)
                      | NONE =>
                            raise Check
                            ("extract_natural_sig: no such module label, "
                            ^ string_of_label lab)
                end
          | _ => raise Error "extract_natural_sig: applied to a non-path"

    (* precondition: con is a path *)
    fun extract_natural_kind ctx con =
        (*
        let val out = TextIO.openAppend "debugging"
            fun prout s = TextIO.output (out, s)
            fun laybnd s f x = Layout.seq [Layout.str s, Layout.str " = ", f x]
            val () = prout "---\nextract_natural_kind:\n"
            val () = Layout.print (laybnd "ctx" Print.IL.ctx_to_layout ctx,
                                   prout)
            val () = prout "\n"
            val () = Layout.print (laybnd "con" Print.IL.con_to_layout con,
                                   prout)
            val () = prout "\n---\n"
            val () = TextIO.closeOut out
        in
        *)
        case con of
            ConVar cx =>
                (case lookup_convar ctx cx of
                    NONE => raise Check
                            ("extract_natural_kind: unbound convar, "
                            ^ Variable.toString cx)
                  | SOME k => k)
          | ConApp (p, c) =>
                (case extract_natural_kind ctx p of
                    KndPi (cx, k, l) => subst_con_kind c cx l
                  | _ => raise Check
                         "extract_natural_kind: function not of Pi kind")
          | ConProj (i, p) =>
                (case extract_natural_kind ctx p of
                    KndProd ks =>
                        (case nth (ks, i-1) of
                            SOME k => k
                          | NONE =>
                                raise Check
                                ("extract_natural_kind: subscript out of bounds, "
                                ^ Int.toString i))
                  | _ => raise Check
                         "extract_natural_kind: projection from non-record")
          | ConModProj (modv, lab) =>
                let val SigStruct sdecs = extract_natural_sig ctx modv
                in
                    case sdecs_lookup_conlab modv sdecs lab of
                        SOME k => k
                      | NONE =>
                            raise Check
                            ("extract_natural_kind: no such module label, "
                            ^ string_of_label lab)
                end
          | _ => raise Error "extract_natural_kind: applied to a non-path"
                 (*
                 before
                 let val out = TextIO.openAppend "debugging"
                     fun prout s = TextIO.output (out, s)
                     val () = prout "\nERROR!\n"
                     val () = Layout.print (Print.IL.con_to_layout con, prout)
                     val () = prout "\nERROR!\n\n"
                     val () = TextIO.closeOut out
                 in
                    ()
                 end
                 *)
        (*
        end
        *)

    fun subkind ctx knd1 knd2 =
        case (knd1, knd2) of
            (KndType, KndType) => ()
          | (KndSing c, KndType) => ()
          | (KndSing c1, KndSing c2) =>
                let val () = equiv_con ctx c1 c2 KndType
                             handle Equiv => raise Subkind
                in
                    ()
                end
          | (KndPi (x1, k1, l1), KndPi (x2, k2, l2)) =>
                let val () = subkind ctx k2 k1
                    val (x, (l1, l2)) = alpha_vary2 (x1, l1) (x2, l2)
                                                    replace_con_kind
                    val ctx = extend ctx (DecConKind (x, k2))
                    val () = subkind ctx l1 l2
                in
                    ()
                end
          | (KndProd knds1, KndProd knds2) =>
                let val () = check (List.length knds1 = List.length knds2)
                             (fn () => raise Subkind)
                    val () =
                        ListPair.app
                            (fn (k1, k2) => subkind ctx k1 k2)
                            (knds1, knds2)
                in
                    ()
                end
          | _ => raise Subkind

    and subsig ctx signat1 signat2 =
        case (signat1, signat2) of
            (SigStruct sdecs1, SigStruct sdecs2) => subsdecs ctx sdecs1 sdecs2

    and subsdecs ctx sdecs1 sdecs2 =
        case (sdecs1, sdecs2) of
            ([], []) => ()
          | ((lab1, d1)::sds1, (lab2, d2)::sds2) =>
                let val () = check (lab1 = lab2) (fn () => raise Subsig)
                    val (d1, (sds1, sds2)) = subdec ctx (d1, sds1) (d2, sds2)
                    val () = subsdecs (extend ctx d1) sds1 sds2
                in
                    ()
                end
          | _ => raise Subsig

    (* does two things in parallel:
        (1) make sure dec1 <: dec2
        (2) alpha_vary2 (dec1, sdecs1) (dec2, sdecs2)
       then returns (dec1', (sdecs1, sdecs2)), where dec1' is the
       alpha-variant of dec1 *)
    and subdec ctx (dec1, sdecs1) (dec2, sdecs2) =
        case (dec1, dec2) of
            (DecExpCon (ex1, c1), DecExpCon (ex2, c2)) =>
                let val (ex, (sdecs1, sdecs2)) = alpha_vary2 (ex1, sdecs1)
                                                             (ex2, sdecs2)
                                                             replace_exp_sdecs
                    val () = equiv_con ctx c1 c2 KndType
                             handle Equiv => raise Subsig
                in
                    (DecExpCon (ex, c1), (sdecs1, sdecs2))
                end
          | (DecConKind (cx1, k1), DecConKind (cx2, k2)) =>
                let val (cx, (sdecs1, sdecs2)) = alpha_vary2 (cx1, sdecs1)
                                                             (cx2, sdecs2)
                                                             replace_con_sdecs
                    val () = subkind ctx k1 k2
                             handle Subkind => raise Subsig
                in
                    (DecConKind (cx, k1), (sdecs1, sdecs2))
                end
          | (DecModSig (mx1, sg1), DecModSig (mx2, sg2)) =>
                let val (mx, (sdecs1, sdecs2)) = alpha_vary2 (mx1, sdecs1)
                                                             (mx2, sdecs2)
                                                             replace_mod_sdecs
                    val () = subsig ctx sg1 sg2
                in
                    (DecModSig (mx, sg1), (sdecs1, sdecs2))
                end
          | _ => raise Subsig

    and whnf ctx con =
        let fun path_reduce ctx conpath =
                case extract_natural_kind ctx conpath of
                    KndSing c => SOME c
                  | _ => NONE
            fun reduce ctx con =
                case con of
                    ConApp (ConAbs (cx, k, c1), c2) =>
                        SOME (subst_con_con c2 cx c1)
                  | ConProj (i, ConProd cs) =>
                        (case nth (cs, i - 1) of
                            SOME c => SOME c
                          | NONE => raise Check("reduce: subscript out of bounds, "
                                                ^ Int.toString i))
                  | ConApp (c1, c2) =>
                        (case reduce ctx c1 of
                            SOME c1' => SOME (ConApp (c1', c2))
                          | NONE => path_reduce ctx con)
                  | ConProj (i, c) =>
                        (case reduce ctx c of
                            SOME c' => SOME (ConProj (i, c'))
                          | NONE => path_reduce ctx con)
                  | ConModProj (modv as ModStruct sbnds, lab) =>
                        (case sbnds_lookup_conlab modv sbnds lab of
                            SOME c => SOME c
                          | NONE => raise Check("reduce: no such module label, "
                                                ^ string_of_label lab))
                  | ConVar _ => path_reduce ctx con
                  | ConEvar ebr =>
                        (case !ebr of
                            EvBound c => SOME c
                          | EvFree _ => NONE)
                  | _ => NONE
        in
            case reduce ctx con of
                SOME con' => whnf ctx con'
              | NONE => con
        end

    and equiv_con ctx con1 con2 knd =
        case knd of
            KndType =>
                let val k = equiv_con_path ctx (whnf ctx con1) (whnf ctx con2)
                    (* is this kind check necessary?  maybe for the sake of
                       belt and suspenders... *)
                    val () = subkind ctx k KndType
                             handle Equiv => raise Error
                             ("mismatch between equiv_con input (KndType) "
                             ^ " and equiv_con_path output")
                in
                    ()
                end
          | KndSing _ => ()
          | KndProd knds =>
                ListUtil.appi
                        (fn (k, i) =>
                            equiv_con ctx (ConProj (i+1, con1))
                                          (ConProj (i+1, con2))
                                          k)
                        knds
          | KndPi (x, k, l) =>
                let val (x, l) = alpha_vary (x, l) replace_con_kind
                in
                    equiv_con (extend ctx (DecConKind (x, k)))
                              (ConApp (con1, ConVar x))
                              (ConApp (con2, ConVar x))
                              l
                end

    and equiv_module_path ctx modp1 modp2 =
        case (modp1, modp2) of
            (ModVar mx1, ModVar mx2) =>
                if Variable.eq (mx1, mx2) then
                    case lookup_modvar ctx mx1 of
                        NONE => raise Check
                                ("equiv_module_path: comparing unbound module variable, "
                                ^ Variable.toString mx1)
                      | SOME sg => sg
                else
                    raise Equiv
          | (ModModProj (mp1, lab1), ModModProj (mp2, lab2)) =>
                let val () = check (label_eq (lab1, lab2))
                                   (fn () => raise Equiv)
                    val SigStruct sdecs = equiv_module_path ctx mp1 mp2
                in
                    case sdecs_lookup_modlab mp1 sdecs lab1 of
                        NONE => raise Check
                               ("equiv_module_path: non-existent module label, "
                               ^ string_of_label lab1)
                      | SOME sg => sg
                end
          | _ => raise Equiv

    and equiv_con_path ctx con1 con2 =
        case (con1, con2) of
            (ConVar cx1, ConVar cx2) =>
                if Variable.eq (cx1, cx2) then
                    case lookup_convar ctx cx1 of
                        NONE => raise Check ("equiv_con_path: comparing unbound variable, "
                                            ^ Variable.toString cx1)
                      | SOME k => k
                else
                    raise Equiv
          | (ConEvar ebr1, c2 as ConEvar ebr2) =>
                if ebr1 = ebr2 then
                    KndType
                else
                    let val () = unify ebr1 c2
                    in
                        KndType
                    end
          | (ConEvar ebr, c) =>
                let val () = check_con_kind ctx c KndType
                             "filling in evar with non-type"
                    val () = unify ebr c
                in  
                    KndType
                end
          | (c, ConEvar ebr) => 
                let val () = check_con_kind ctx c KndType
                             "filling in evar with non-type"
                    val () = unify ebr c
                in  
                    KndType
                end
          | (ConApp (cp1, c1), ConApp (cp2, c2)) =>
                let val k = equiv_con_path ctx cp1 cp2
                in
                    case k of
                        KndPi (x, k1, k2) =>
                            let val () = equiv_con ctx c1 c2 k1
                            in
                                subst_con_kind c1 x k2
                            end
                      | _ => raise Check
                             "equiv_con_path: application of non-function"
                end
          | (ConProj (i1, cp1), ConProj (i2, cp2)) =>
                let val () = check (i1 = i2) (fn () => raise Equiv)
                in
                    case equiv_con_path ctx cp1 cp2 of
                        KndProd knds =>
                           (case nth (knds, i1 - 1) of
                               SOME k => k
                             | NONE =>
                                raise Check
                                ("equiv_con_path: subscript out of bounds, "
                                ^ Int.toString i1))
                      | _ => raise Check
                             "equiv_con_path: projection from a non-tuple"
                end
          | (ConModProj (modp1, lab1), ConModProj (modp2, lab2)) =>
                let val () = check (label_eq (lab1, lab2))
                                   (fn () => raise Equiv)
                    val SigStruct sdecs = equiv_module_path ctx modp1 modp2
                in
                    case sdecs_lookup_conlab modp1 sdecs lab1 of
                        NONE => raise Check
                                ("equiv_con_path: non-existent module label, "
                                ^ string_of_label lab1)
                      | SOME k => k
                end

          | (TyBool, TyBool) => KndType
          | (TyUnit, TyUnit) => KndType
          | (TyLit l, TyLit l') =>
              if Literal.lit_ty_eq (l, l')
              then KndType
              else raise Check "equiv_con_path: literal types don't match"
          | (TyProd (c1left, c1right), TyProd (c2left, c2right)) =>
                let val () = equiv_con ctx c1left c2left KndType
                    val () = equiv_con ctx c1right c2right KndType
                in
                    KndType
                end
          | (TyList c1, TyList c2) =>
                let val () = equiv_con ctx c1 c2 KndType
                in
                    KndType
                end
          | (TyArrow (c1dom, c1cod), TyArrow (c2dom, c2cod)) =>
                let val () = equiv_con ctx c1dom c2dom KndType
                    val () = equiv_con ctx c1cod c2cod KndType
                in
                    KndType
                end
          | (TyForall (cx1, k1, c1), TyForall (cx2, k2, c2)) =>
                let val () = equiv_kind ctx k1 k2
                    val (cx, (c1, c2)) = alpha_vary2 (cx1, c1) (cx2, c2)
                                                     replace_con_con
                    val () = equiv_con (extend ctx (DecConKind (cx, k1)))
                                       c1 c2 KndType
                 in
                     KndType
                 end

          | (ConAbs _, ConAbs _) => raise Error "equiv_con_path on non-path"
          | _ => raise Equiv


    and check_equiv_con ctx con1 con2 knd s =
        equiv_con ctx con1 con2 knd
        handle Equiv => raise Check s

    and equiv_kind ctx knd1 knd2 =
        case (knd1, knd2) of
            (KndType, KndType) => ()
          | (KndPi (cx1, k1, l1), KndPi (cx2, k2, l2)) =>
                let val () = equiv_kind ctx k1 k2
                    val (cx, (l1, l2)) = alpha_vary2 (cx1, l1) (cx2, l2)
                                                     replace_con_kind
                    val ctx = extend ctx (DecConKind (cx, k1))
                    val () = equiv_kind ctx l1 l2
                in
                    ()
                end
          | (KndProd knds1, KndProd knds2) =>
                let val () = check (List.length knds1 = List.length knds2)
                                   (fn () => raise Equiv)
                    val () = ListPair.app (fn (k1, k2) => equiv_kind ctx k1 k2)
                                          (knds1, knds2)
                in
                    ()
                end
          | (KndSing c1, KndSing c2) => equiv_con ctx c1 c2 KndType
          | _ => raise Equiv

    and check_kind ctx knd =
        case knd of
            KndType => ()
          | KndPi (cx, k, l) =>
                let val () = check_kind ctx k
                    val (cx, l) = alpha_vary (cx, l) replace_con_kind
                    val ctx = extend ctx (DecConKind (cx, k))
                    val () = check_kind ctx l
                in
                    ()
                end
          | KndProd ks => List.app (fn k => check_kind ctx k) ks
          | KndSing c => check_con_kind ctx c KndType
                         "singleton at kind other than Type"

    and synth_mod_sig ctx module =
        case module of
            ModVar mx =>
                (case lookup_modvar ctx mx of
                    SOME (SigStruct sdecs) => SigStruct (selfify module sdecs)
                  | NONE => raise Check ("synth_mod_sig: unbound modvar, " ^ Variable.toString mx))
          | ModStruct sbnds => SigStruct (synth_sbnds_sdecs ctx sbnds)
          | ModModProj (modv, lab) =>
                let val SigStruct sdecs = synth_mod_sig ctx modv
                    (* belt and suspenders check *)
                    val () = if module_value modv then
                                ()
                             else
                                raise Check "synth_mod_sig: projection from a non-value module"
                    val sdecs = selfify modv sdecs
                in
                    case sdecs_lookup_modlab modv sdecs lab of
                        NONE => raise Check ("synth_mod_sig: no such module label, "
                                            ^ string_of_label lab)
                      | SOME sg => sg
                end
          | ModConstrain (m, sg) =>
                let val () = check_mod_sig ctx m sg
                             "synth_mod_sig: ascribed module doesn't match ascribed signature"
                in
                    sg
                end

    and check_mod_sig ctx module signat s =
        let val signat' = synth_mod_sig ctx module
        in
            subsig ctx signat' signat
            handle Subsig => raise Check s
        end

    and synth_sbnds_sdecs ctx sbnds =
        case sbnds of
            [] => []
          | (lab, bnd)::sbs =>
                let val dec = 
                    case bnd of
                        BndExp (ex, e) => DecExpCon (ex, synth_exp_con ctx e)
                      | BndCon (cx, c) => DecConKind (cx, synth_con_kind ctx c)
                      | BndMod (mx, m) => DecModSig (mx, synth_mod_sig ctx m)
                in
                    (lab, dec) :: synth_sbnds_sdecs (extend ctx dec) sbs
                end

    and synth_con_kind ctx con =
        case con of
            ConVar cx => 
                (case lookup_convar ctx cx of
                    SOME k => labelledKndSing con k
                  | NONE => raise Check ("synth_con_kind: unbound convar, " ^ Variable.toString cx))
          | ConEvar ebr =>
                (case !ebr of
                    EvBound c => synth_con_kind ctx c
                  | EvFree _ => KndSing con)
          | ConAbs (cx, k, c) =>
                let val () = check_kind ctx k
                    val (cx, c) = alpha_vary (cx, c) replace_con_con
                    val ctx = extend ctx (DecConKind (cx, k))
                    val k' = synth_con_kind ctx c
                in
                    KndPi (cx, k, k')
                end
          | ConApp (c1, c2) =>
                let val k1 = synth_con_kind ctx c1
                in
                    case k1 of
                        KndPi (cx, k, k') =>
                            let val () = check_con_kind ctx c2 k
                                         "synth_con_kind: argument in ConApp has wrong kind"
                            in
                                subst_con_kind c2 cx k'
                            end
                      | _ => raise Check "synth_con_kind: function in ConApp has wrong kind"
                end
          | ConProd cons => KndProd (map (fn c => synth_con_kind ctx c) cons)
          | ConProj (i, c) =>
                let val k = synth_con_kind ctx c
                in
                    case k of
                        KndProd ks =>
                            (case nth (ks, i - 1) of
                                SOME k => k
                              | NONE => raise Check
                                    ("synth_con_kind: subscript out of bounds, "
                                    ^ Int.toString i))
                      | _ => raise Check ("synth_con_kind: projection from a non-tuple con")
                end
          | ConModProj (modv, lab) =>
                let val SigStruct sdecs = synth_mod_sig ctx modv
                    (* belt and suspenders check *)
                    val () = if module_value modv then
                                ()
                             else
                                raise Check "synth_con_kind: projection from a non-value module"
                    val sdecs = selfify modv sdecs
                in
                    case sdecs_lookup_conlab modv sdecs lab of
                        NONE => raise Check ("synth_con_kind: no such module label, "
                                            ^ string_of_label lab)
                      | SOME k => k
                end
          | TyBool => KndSing con
          | TyLit l => KndSing con
          | TyUnit => KndSing con
          | TyProd (c1, c2) =>
                let val () = check_con_kind ctx c1 KndType 
                            "synth_con_kind: TyProd with non-type component"
                    val () = check_con_kind ctx c2 KndType
                            "synth_con_kind: TyProd with non-type component"
                in
                    KndSing con
                end
          | TyList c =>
                let val () = check_con_kind ctx c KndType
                             "synth_con_kind: TyList with non-type argument"
                in
                    KndSing con
                end
          | TyArrow (c1, c2) =>
                let val () = check_con_kind ctx c1 KndType 
                            "synth_con_kind: TyArrow with non-type component"
                    val () = check_con_kind ctx c2 KndType
                            "synth_con_kind: TyArrow with non-type component"
                in
                    KndSing con
                end
          | TyForall (cx, k, c) =>
                let val () = check_kind ctx k
                    val (cx, c) = alpha_vary (cx, c) replace_con_con
                    val ctx = extend ctx (DecConKind (cx, k))
                    val () = check_con_kind ctx c KndType
                             "synth_con_kind: TyForall with non-type body"
                in
                    KndSing con
                end

    and check_con_kind ctx con knd s =
        let val knd' = synth_con_kind ctx con
        in
            subkind ctx knd' knd
            handle Subkind => raise Check s
        end

    and synth_exp_con ctx exp =
        case exp of
            ExVar x => 
                (case lookup_expvar ctx x of
                    SOME ty => ty
                  | NONE => raise Check ("synth_exp_con: unbound expvar, " ^ Variable.toString x))
          | ExUnit => TyUnit
          | ExBool _ => TyBool
          | ExIf (e1, e2, e3) =>
                let val () = check_exp_con ctx e1 TyBool
                             "synth_exp_con: if guard not of type bool"
                    val ty2 = synth_exp_con ctx e2
                    val ty3 = synth_exp_con ctx e3
                    val () = check_equiv_con ctx ty2 ty3 KndType
                             "synth_exp_con: if arms have inequivalent types"
                in
                    ty2
                end
          | ExLit l => TyLit (Literal.lit_ty l)
          | ExUnOp (oper, e) =>
                let val (cin, cout) =
                          Literal.unop_tys oper TyLit TyBool TyUnit
                    val c = synth_exp_con ctx e
                    val () = check_equiv_con ctx cin c KndType
                             ("synth_exp_con: argument of operator " ^
                             "expression incorrect")
                in
                  cout
                end
          | ExBinOp (e1, oper, e2) =>
                let val (cin1, cin2, cout) =
                          Literal.binop_tys oper TyLit TyBool TyUnit
                    val c1 = synth_exp_con ctx e1
                    val () = check_equiv_con ctx cin1 c1 KndType
                             ("synth_exp_con: left argument of operator " ^
                             "expression incorrect")
                    val c2 = synth_exp_con ctx e2
                    val () = check_equiv_con ctx cin2 c2 KndType
                             ("synth_exp_con: right argument of operator " ^
                             "expression incorrect")
                in
                  cout
                end
          | ExPair (e1, e2) =>
                let val ty1 = synth_exp_con ctx e1
                    val ty2 = synth_exp_con ctx e2
                in
                    TyProd (ty1, ty2)
                end
          (* XXX should this be evar-sensitive? *)
          | ExProj1 e =>
                let val c = synth_exp_con ctx e
                in
                    case whnf ctx c of
                        TyProd (c1, _) => c1
                      | _ => raise Check ("synth_exp_con: #1 from non-pair")
                end
          | ExProj2 e =>
                let val c = synth_exp_con ctx e
                in
                    case whnf ctx c of
                        TyProd (_, c2) => c2
                      | _ => raise Check ("synth_exp_con: #2 from non-pair")
                end
          | ExNil c => TyList c
          | ExCons (e1, e2) =>
                let val c1 = synth_exp_con ctx e1
                    val c2 = synth_exp_con ctx e2
                in
                    case whnf ctx c2 of
                        TyList c =>
                            let val () = check_equiv_con ctx c1 c KndType
                                         "synth_exp_con: incompatible cons"
                            in
                                c2
                            end
                      | _ => raise Check "synth_exp_con: cons onto non-list"
                end
          | ExCase (e, e_nil, (x, xs, e_cons)) =>
                let val c = synth_exp_con ctx e
                in
                    case whnf ctx c of
                        TyList c' =>
                            let val c_nil = synth_exp_con ctx e_nil
                                val ctx = extend_many ctx
                                            [DecExpCon (x, c'),
                                             DecExpCon (xs, TyList c')]
                                val c_cons = synth_exp_con ctx e_cons
                                val () = check_equiv_con ctx c_nil c_cons KndType
                                         "synth_exp_con: case arms have different type"
                            in
                                c_nil
                            end
                      | _ => raise Check "synth_exp_con: case over non-list"
                end
          | ExFix (f, x, c1, c2, e) =>
                let val ctx = extend_many ctx [DecExpCon (f, TyArrow (c1, c2)),
                                               DecExpCon (x, c1)]
                    val c2' = synth_exp_con ctx e
                    val () = check_equiv_con ctx c2 c2' KndType
                             "synth_exp_con: incorrect body type in fix expression"
                in
                    TyArrow (c1, c2)
                end
          | ExApp (e1, e2) =>
                let val c1 = synth_exp_con ctx e1
                in
                    case whnf ctx c1 of
                        TyArrow (c1dom, c1cod) =>
                            let val c2 = synth_exp_con ctx e2
                                val () = check_equiv_con ctx c1dom c2 KndType
                                         "synth_exp_con: function argument doesn't match function domain"
                            in
                                c1cod
                            end
                      | _ => raise Check "synth_exp_con: application of a non-function"
                end
          | ExConAbs (cx, k, e) =>
                let val () = check_kind ctx k
                    val (cx, e) = alpha_vary (cx, e) replace_con_exp
                    val ctx = extend ctx (DecConKind (cx, k))
                    val c = synth_exp_con ctx e
                in
                    TyForall (cx, k, c)
                end
          | ExConApp (e, c_arg) =>
                let val c_fun = synth_exp_con ctx e
                in
                    case whnf ctx c_fun of
                        TyForall (cx, k, c) => subst_con_con c_arg cx c
                      | _ => raise Check "synth_exp_con: con-application of a non-forall"
                end
          | ExModLet (mx, m, e) =>
                let val sg = synth_mod_sig ctx m
                    val (mx, e) = alpha_vary (mx, e) replace_mod_exp
                    val ctx = extend ctx (DecModSig (mx, sg))
                    val c = synth_exp_con ctx e
                    (* <<normalize>> *)
                    val () = if ModVarSet.member (fmv_con c, mx) then
                                raise Check "synth_exp_con: module let-binding can't be given non-dependent type"
                             else
                                ()
                in
                    c
                end
          | ExModProj (modv, lab) =>
                let val SigStruct sdecs = synth_mod_sig ctx modv
                    (* belt and suspenders check *)
                    val () = if module_value modv then
                                ()
                             else
                                raise Check "synth_exp_con: projection from a non-value module"
                    val sdecs = selfify modv sdecs
                in
                    case sdecs_lookup_explab modv sdecs lab of
                        NONE => raise Check ("synth_exp_con: no such module label, "
                                            ^ string_of_label lab)
                      | SOME c => c
                end

    and check_exp_con ctx exp con s =
        let val con' = synth_exp_con ctx exp
        in
            equiv_con ctx con con' KndType
            handle Equiv => raise Check s
        end
    (* Inference helper functions *)
    structure Inference :> INFERENCE =
      struct
        type tcontext = tcontext
        type context = context

        exception Occurs = Occurs
        exception Equiv = Equiv

        fun accessible_convars ctx =
            let val convars = List.mapPartial
                                (fn (_, DecConKind (cx, _)) => SOME cx
                                  | _ => NONE)
                                ctx
            in
                ConVarSet.addList (ConVarSet.empty, convars)
            end

        fun accessible_modvars ctx =
            let val modvars = List.mapPartial
                                (fn (_, DecModSig (cx, _)) => SOME cx
                                  | _ => NONE)
                                ctx
            in
                ModVarSet.addList (ModVarSet.empty, modvars)
            end

        fun new_con ctx =
            let val condeps = accessible_convars ctx
                val moddeps = accessible_modvars ctx
            in
                ConEvar (ref (EvFree (condeps, moddeps)))
            end

        fun instantiate ctx (e, c) =
          let
            val erase = map (fn (_, dec) => dec)
            fun instantiator ctx knd =
                let val ks = case knd of
                                KndProd lks => lks
                              | _ => raise Error "forall over non-product kind"
                    val cs = List.map (fn k => new_con ctx) ks
                in
                    ConProd cs
                end
          in
            case whnf (erase ctx) c of
                TyForall (cx, k, c') =>
                    let val cp = instantiator ctx k
                    in
                        (ExConApp (e, cp), subst_con_con cp cx c')
                    end
              | c => (e, c)
          end


        fun generalize ctx checker =
          let
            fun generalizer (exp, con) convar =
                let val n = ref 0
                    fun gen_con con =
                        map_con (fn cx => ConVar cx)
                                (fn ebr =>
                                    let val () =
                                        case !ebr of
                                            EvBound con => (gen_con con; ())
                                          | EvFree (condeps, moddeps) =>
                                                (* all the work goes on here *)
                                                if ConVarSet.member
                                                    (condeps, convar)
                                                then
                                                    ( (* pre-increment *)
                                                      n := !n + 1 ;
                                                      ebr :=
                                                          EvBound
                                                          (ConProj 
                                                            (!n, ConVar convar))
                                                    )
                                                else
                                                    ()
                                    in
                                        ConEvar ebr
                                    end)
                                (fn (cx, c) => (cx, gen_con c))
                                gen_kind
                                gen_module
                                con
                    and gen_kind knd =
                        map_kind (fn (cx, k) => (cx, gen_kind k))
                                 gen_con
                                 knd
                    and gen_module module =
                        map_module (fn mx => ModVar mx)
                                   gen_sbnds
                                   gen_signat
                                   module
                    and gen_sbnds sbnds =
                        map_sbnds (fn (ex, sbs) => (ex, gen_sbnds sbs))
                                  (fn (cx, sbs) => (cx, gen_sbnds sbs))
                                  (fn (mx, sbs) => (mx, gen_sbnds sbs))
                                  gen_exp
                                  gen_con
                                  gen_module
                                  sbnds
                    and gen_signat signat =
                        map_signat gen_sdecs
                                   signat
                    and gen_sdecs sdecs =
                        map_sdecs (fn (ex, sds) => (ex, gen_sdecs sds))
                                  (fn (cx, sds) => (cx, gen_sdecs sds))
                                  (fn (mx, sds) => (mx, gen_sdecs sds))
                                  gen_con
                                  gen_kind
                                  gen_signat
                                  sdecs
                    and gen_exp exp =
                        map_exp (fn ex => ExVar ex)
                                (fn (cx, e) => (cx, gen_exp e))
                                (fn (mx, e) => (mx, gen_exp e))
                                gen_con
                                gen_kind
                                gen_module
                                exp
                in
                    gen_con con;
                    if !n = 0 then
                        (* no generalization *)
                        (exp, con)
                    else
                        let val k = kndExp (KndType, !n)
                        in
                            (ExConAbs (convar, k, exp),
                             TyForall (convar, k, con))
                        end
                end
            val cx_scope = Variable.newNamed "a"
            val ctx = extend ctx (LabInternal "scope",
                                  DecConKind (cx_scope, KndType))
          in
            generalizer (checker ctx) cx_scope
          end

        val unify = equiv_con
      end
  end
